@@ -1,11 +1,15 @@
 """This module provides methods for handling queries."""
 import re
-from typing import List, Dict, Set, Optional
-from uvicorn.config import logger
-from disease import NAMESPACE_LOOKUP, PREFIX_LOOKUP, SOURCES_LOWER_LOOKUP
+from typing import Dict, Set, Optional
+from .version import __version__
+from disease import NAMESPACE_LOOKUP, PREFIX_LOOKUP, SOURCES_LOWER_LOOKUP,\
+    ITEM_TYPES, logger
 from disease.database import Database
-from disease.schemas import Disease, Meta, MatchType, SourceName
+from disease.schemas import Disease, SourceMeta, MatchType, SourceName, \
+    ServiceMeta, NormalizationService, SearchService
 from botocore.exceptions import ClientError
+from urllib.parse import quote
+from datetime import datetime
 
 
 class InvalidParameterException(Exception):
@@ -13,7 +17,6 @@ class InvalidParameterException(Exception):
 
     def __init__(self, message):
         """Create new instance
-
         :param str message: string describing the nature of the error
         """
         super().__init__(message)
@@ -26,7 +29,6 @@ class QueryHandler:
 
     def __init__(self, db_url: str = '', db_region: str = 'us-east-2'):
         """Initialize Normalizer instance.
-
         :param str db_url: URL to database source.
         :param str db_region: AWS default region.
         """
@@ -34,7 +36,6 @@ class QueryHandler:
 
     def _emit_warnings(self, query_str) -> Optional[Dict]:
         """Emit warnings if query contains non breaking space characters.
-
         :param str query_str: query string
         :return: dict keying warning type to warning description
         """
@@ -49,9 +50,8 @@ class QueryHandler:
             )
         return warnings
 
-    def _fetch_meta(self, src_name: str) -> Meta:
+    def _fetch_meta(self, src_name: str) -> SourceMeta:
         """Fetch metadata for src_name.
-
         :param str src_name: name of source to get metadata for
         :return: Meta object containing source metadata
         """
@@ -62,7 +62,7 @@ class QueryHandler:
                 db_response = self.db.metadata.get_item(
                     Key={'src_name': src_name}
                 )
-                response = Meta(**db_response['Item'])
+                response = SourceMeta(**db_response['Item'])
                 self.db.cached_sources[src_name] = response
                 return response
             except ClientError as e:
@@ -73,20 +73,14 @@ class QueryHandler:
                     item: Dict,
                     match_type: str) -> (Dict, str):
         """Add individual record to response object
-
         :param Dict[str, Dict] response: in-progress response object to return
             to client
         :param Dict item: Item retrieved from DynamoDB
-        :param MatchType match_type: type of query match
+        :param str match_type: type of query match
         :return: Tuple containing updated response object, and string
             containing name of the source of the match
         """
         del item['label_and_type']
-        attr_types = ['aliases', 'other_identifiers', 'xrefs']
-        for attr_type in attr_types:
-            if attr_type not in item.keys():
-                item[attr_type] = []
-
         disease = Disease(**item)
         src_name = item['src_name']
 
@@ -97,12 +91,12 @@ class QueryHandler:
             matches[src_name] = {
                 'match_type': MatchType[match_type.upper()],
                 'records': [disease],
-                'meta_': self._fetch_meta(src_name)
+                'source_meta_': self._fetch_meta(src_name)
             }
         elif matches[src_name]['match_type'] == MatchType[match_type.upper()]:
             matches[src_name]['records'].append(disease)
 
-        return (response, src_name)
+        return response, src_name
 
     def _fetch_records(self,
                        response: Dict[str, Dict],
@@ -110,7 +104,6 @@ class QueryHandler:
                        match_type: str) -> (Dict, Set):
         """Return matched Disease records as a structured response for a given
         collection of concept IDs.
-
         :param Dict[str, Dict] response: in-progress response object to return
             to client.
         :param List[str] concept_ids: List of concept IDs to build from.
@@ -129,11 +122,10 @@ class QueryHandler:
                 matched_sources.add(src)
             except ClientError as e:
                 logger.error(e.response['Error']['Message'])
-        return (response, matched_sources)
+        return response, matched_sources
 
     def _fill_no_matches(self, resp: Dict[str, Dict]) -> Dict:
         """Fill all empty source_matches slots with NO_MATCH results.
-
         :param Dict[str, Dict] resp: incoming response object
         :return: response object with empty source slots filled with
                 NO_MATCH results and corresponding source metadata
@@ -143,7 +135,7 @@ class QueryHandler:
                 resp['source_matches'][src_name] = {
                     'match_type': MatchType.NO_MATCH,
                     'records': [],
-                    'meta_': self._fetch_meta(src_name)
+                    'source_meta_': self._fetch_meta(src_name)
                 }
         return resp
 
@@ -152,7 +144,6 @@ class QueryHandler:
                           resp: Dict,
                           sources: Set[str]) -> (Dict, Set):
         """Check query for concept ID match. Should only find 0 or 1 matches.
-
         :param str query: search string
         :param Dict resp: in-progress response object to return to client
         :param Set[str] sources: remaining unmatched sources
@@ -174,7 +165,7 @@ class QueryHandler:
             (resp, src_name) = self._add_record(resp, item,
                                                 MatchType.CONCEPT_ID.name)
             sources = sources - {src_name}
-        return (resp, sources)
+        return resp, sources
 
     def _check_match_type(self,
                           query: str,
@@ -186,7 +177,7 @@ class QueryHandler:
         :param Dict resp: in-progress response object to return to client
         :param Set[str] sources: remaining unmatched sources
         :param str match_type: Match type to check for. Should be one of
-            {'label', 'alias'}
+            {'label', 'alias', 'xref', 'associated_with'}
         :return: Tuple with updated resp object and updated set of unmatched
                  sources
         """
@@ -196,12 +187,11 @@ class QueryHandler:
             (resp, matched_srcs) = self._fetch_records(resp, concept_ids,
                                                        match_type)
             sources = sources - matched_srcs
-        return (resp, sources)
+        return resp, sources
 
     def _response_keyed(self, query: str, sources: Set[str]) -> Dict:
         """Return response as dict where key is source name and value
         is a list of records. Corresponds to `keyed=true` API parameter.
-
         :param str query: string to match against
         :param Set[str] sources: sources to match from
         :return: completed response object to return to client
@@ -223,8 +213,7 @@ class QueryHandler:
         if len(sources) == 0:
             return response
 
-        match_types = ['label', 'trade_name', 'alias']
-        for match_type in match_types:
+        for match_type in ITEM_TYPES.values():
             (response, sources) = self._check_match_type(query, response,
                                                          sources, match_type)
             if len(sources) == 0:
@@ -233,12 +222,11 @@ class QueryHandler:
         # remaining sources get no match
         return self._fill_no_matches(response)
 
-    def _response_list(self, query: str, sources: List[str]) -> Dict:
+    def _response_list(self, query: str, sources: Set[str]) -> Dict:
         """Return response as list, where the first key-value in each item
         is the source name. Corresponds to `keyed=false` API parameter.
-
         :param str query: string to match against
-        :param List[str] sources: sources to match from
+        :param Set[str] sources: sources to match from
         :return: Completed response object to return to client
         """
         response_dict = self._response_keyed(query, sources)
@@ -257,7 +245,6 @@ class QueryHandler:
 
     def search_sources(self, query_str, keyed=False, incl='', excl='') -> Dict:
         """Fetch normalized disease objects.
-
         :param str query_str: query, a string, to search for
         :param bool keyed: if true, return response as dict keying source names
             to source objects; otherwise, return list of source objects
@@ -270,8 +257,12 @@ class QueryHandler:
         :raises InvalidParameterException: if both incl and excl args are
             provided, or if invalid source names are given.
         """
+        sources = dict()
+        for k, v in SOURCES_LOWER_LOOKUP.items():
+            if self.db.metadata.get_item(Key={'src_name': v}).get('Item'):
+                sources[k] = v
         if not incl and not excl:
-            query_sources = set(SOURCES_LOWER_LOOKUP.values())
+            query_sources = set(sources.values())
         elif incl and excl:
             detail = "Cannot request both source inclusions and exclusions."
             raise InvalidParameterException(detail)
@@ -280,8 +271,8 @@ class QueryHandler:
             invalid_sources = []
             query_sources = set()
             for source in req_sources:
-                if source.lower() in SOURCES_LOWER_LOOKUP.keys():
-                    query_sources.add(SOURCES_LOWER_LOOKUP[source.lower()])
+                if source.lower() in sources.keys():
+                    query_sources.add(sources[source.lower()])
                 else:
                     invalid_sources.append(source)
             if invalid_sources:
@@ -293,9 +284,9 @@ class QueryHandler:
             invalid_sources = []
             query_sources = set()
             for req_l, req in req_excl_dict.items():
-                if req_l not in SOURCES_LOWER_LOOKUP.keys():
+                if req_l not in sources.keys():
                     invalid_sources.append(req)
-            for src_l, src in SOURCES_LOWER_LOOKUP.items():
+            for src_l, src in sources.items():
                 if src_l not in req_excl_dict.keys():
                     query_sources.add(src)
             if invalid_sources:
@@ -309,104 +300,155 @@ class QueryHandler:
         else:
             response = self._response_list(query_str, query_sources)
 
-        return response
+        response['service_meta_'] = ServiceMeta(
+            version=__version__,
+            response_datetime=datetime.now(),
+        ).dict()
+        return SearchService(**response).dict()
 
     def _add_merged_meta(self, response: Dict) -> Dict:
         """Add source metadata to response object.
+
         :param Dict response: in-progress response object
-        :return: completed resopnse object.
+        :return: completed response object.
         """
         sources_meta = {}
-        for concept_id in response['record']['concept_ids']:
+        vod = response['disease_descriptor']
+        ids = [vod['disease_id']] + vod.get('xrefs', [])
+        for concept_id in ids:
             prefix = concept_id.split(':')[0]
             src_name = PREFIX_LOOKUP[prefix.lower()]
             if src_name not in sources_meta:
                 sources_meta[src_name] = self._fetch_meta(src_name)
-        response['meta_'] = sources_meta
+        response['source_meta_'] = sources_meta
         return response
 
-    def _format_merged_record(self, merged_record: Dict) -> Dict:
-        """Transform merged record (as served by DB) into response object
-        format.
-        :param Dict merged_record: merged record as returned by DB call
-        :return: formatted record
-        """
-        del merged_record['label_and_type']
-        merged_record['concept_ids'] = merged_record['concept_id'].split('|')
-        del merged_record['concept_id']
-        return merged_record
+    def _add_vod(self, response: Dict, record: Dict, query: str,
+                 match_type: MatchType) -> Dict:
+        """Format received DB record as VOD and update response object.
 
-    def _add_merged_record(self, response: Dict, merge_ref: str) -> Dict:
-        """Add referenced concept ID group to response object.
-        :param Dict response: in-progress response object. Should have
-            match_type attribute filled in already.
-        :param str merge_ref: primary key for concept ID group lookup.
-        :return: completed response object.
+        :param Dict response: in-progress response object
+        :param Dict record: record as stored in DB
+        :param str query: query string from user request
+        :param MatchType match_type: type of match achieved
+        :return: completed response object ready to return to user
         """
-        merged_record = self.db.get_record_by_id(merge_ref, False, True)
-        if not merged_record:
-            logger.warning(f"Could not retrieve merged record for concept "
-                           f"ID group {merge_ref} "
-                           f"by way of query {response['query']}")
-            response['match_type'] = MatchType.NO_MATCH
-            return response
-        response['record'] = self._format_merged_record(merged_record)
+        vod = {
+            'id': f'normalize.disease:{quote(query)}',
+            'type': 'DiseaseDescriptor',
+            'disease_id': record['concept_id'],
+            'label': record['label'],
+            'extensions': [],
+        }
+        if 'xrefs' in record:
+            vod['xrefs'] = record['xrefs']
+        if 'aliases' in record:
+            vod['alternate_labels'] = record['aliases']
+        if 'pediatric_disease' in record and \
+                record['pediatric_disease'] is not None:
+            vod['extensions'].append({
+                'type': 'Extension',
+                'name': 'pediatric_disease',
+                'value': record['pediatric_disease']
+            })
+        if 'associated_with' in record and record['associated_with']:
+            vod['extensions'].append({
+                'type': 'Extension',
+                'name': 'associated_with',
+                'value': record['associated_with']
+            })
+        if not vod['extensions']:
+            del vod['extensions']
+        response['match_type'] = match_type
+        response['disease_descriptor'] = vod
         response = self._add_merged_meta(response)
         return response
 
     def _record_order(self, record: Dict) -> (int, str):
         """Construct priority order for matching. Only called by sort().
+
         :param Dict record: individual record item in iterable to sort
         :return: tuple with rank value and concept ID
         """
-        # TODO reevaluate/update
         src = record['src_name']
         if src == SourceName.NCIT.value:
             source_rank = 1
         elif src == SourceName.MONDO.value:
             source_rank = 2
-        elif src == SourceName.DO.value:
+        elif src == SourceName.ONCOTREE.value:
             source_rank = 3
+        elif src == SourceName.OMIM.value:
+            source_rank = 4
+        elif src == SourceName.DO.value:
+            source_rank = 5
         else:
             logger.warning(f"query.record_order: Invalid source name for "
                            f"{record}")
-        return (source_rank, record['concept_id'])
+            source_rank = 4
+        return source_rank, record['concept_id']
 
-    def search_groups(self, query_str: str) -> Dict:
-        """Return merged, normalized concept for given search term.
-        :param str query_str: string to search against
+    def _handle_failed_merge_ref(self, record, response, query) -> Dict:
+        """Log + fill out response for a failed merge reference lookup.
+
+        :param Dict record: record containing failed merge_ref
+        :param Dict response: in-progress response object
+        :param str query: original query value
+        :return: response with no match
+        """
+        logger.error(f"Merge ref lookup failed for ref {record['merge_ref']} "
+                     f"in record {record['concept_id']} from query {query}")
+        response['match_type'] = MatchType.NO_MATCH
+        return response
+
+    def search_groups(self, query: str) -> Dict:
+        """Return normalized concept for given search term.
+        :param str query: string to search against
         """
         # prepare basic response
         response = {
-            'query': query_str,
-            'warnings': self._emit_warnings(query_str),
+            'query': query,
+            'warnings': self._emit_warnings(query),
+            'service_meta_': ServiceMeta(
+                version=__version__,
+                response_datetime=datetime.now(),
+            ).dict()
         }
-        if query_str == '':
+        if query == '':
             response['match_type'] = MatchType.NO_MATCH
             return response
-        query_str = query_str.lower()
+        query_str = query.lower().strip()
 
         # check merged concept ID match
         record = self.db.get_record_by_id(query_str, case_sensitive=False,
                                           merge=True)
         if record:
-            response['match_type'] = MatchType.CONCEPT_ID
-            response['record'] = self._format_merged_record(record)
-            response = self._add_merged_meta(response)
-            return response
+            return self._add_vod(response, record, query,
+                                 MatchType.CONCEPT_ID)
+
+        non_merged_match = None
 
         # check concept ID match
         record = self.db.get_record_by_id(query_str, case_sensitive=False)
         if record:
-            response['match_type'] = MatchType.CONCEPT_ID
-            response = self._add_merged_record(response, record['merge_ref'])
-            return response
+            if 'merge_ref' in record:
+                merge = self.db.get_record_by_id(record['merge_ref'],
+                                                 case_sensitive=False,
+                                                 merge=True)
+                if merge is None:
+                    return self._handle_failed_merge_ref(record, response,
+                                                         query_str)
+                else:
+                    return self._add_vod(response, merge, query,
+                                         MatchType.CONCEPT_ID)
+            else:
+                non_merged_match = (record, 'concept_id')
 
         # check other match types
-        for match_type in ['label', 'alias']:
+        for match_type in ['label', 'alias', 'xref', 'associated_with']:
             # get matches list for match tier
             query_matches = self.db.get_records_by_type(query_str, match_type)
-            query_matches = [self.db.get_record_by_id(m['concept_id'], False)
+            query_matches = [self.db.get_record_by_id(m['concept_id'],
+                                                      case_sensitive=False)
                              for m in query_matches]
             query_matches.sort(key=self._record_order)
 
@@ -414,11 +456,28 @@ class QueryHandler:
             for match in query_matches:
                 record = self.db.get_record_by_id(match['concept_id'], False)
                 if record:
-                    merge_ref = record['merge_ref']
-                    if merge_ref:
-                        response['match_type'] = MatchType[match_type.upper()]
-                        return self._add_merged_record(response, merge_ref)
+                    if 'merge_ref' in record:
+                        merge = self.db.get_record_by_id(record['merge_ref'],
+                                                         case_sensitive=False,
+                                                         merge=True)
+                        if merge is None:
+                            return self._handle_failed_merge_ref(record,
+                                                                 response,
+                                                                 query_str)
+                        else:
+                            return self._add_vod(response, merge, query,
+                                                 MatchType[match_type.upper()])
+                    else:
+                        if not non_merged_match:
+                            non_merged_match = (record, match_type)
+
+        # if no successful match, try available non-merged match
+        if non_merged_match:
+            match_type = MatchType[non_merged_match[1].upper()]
+            response = self._add_vod(response, non_merged_match[0], query,
+                                     match_type)
+            return response
 
         if not query_matches:
             response['match_type'] = MatchType.NO_MATCH
-        return response
+        return NormalizationService(**response).dict()
